from schematics import Model, types as t

class Error(Model):

    error_code = t.StringType(required=True)
    error_name = t.StringType(required=True)
    message = t.StringType(required=True)

    def set_format_args(self, *args):
        if args:
            self.message = self.message.format(*args)


from ..constants import *
from ..objects import *
from ..errors import *


def get_data_mapping(feature: Feature, handler: FeatureHandler) -> str:

    # Use the data mapping from the feature handler.
    data_mapping = handler.data_mapping
    
    # Use the feature data mapping if there is no feature handler data mapping.
    if not data_mapping:
        data_mapping = feature.data_mapping

    # Use the feature group data mapping if there is no feature or feature handler data mapping.
    if not data_mapping:
        data_mapping = feature.group.data_mapping
    
    # Return data mapping.
    return data_mapping


from ...data import *


cache: Dict[str, FeatureGroupData] = {}


def get(id: str):
    group_name, feature_name = id.split('.')
    group_data = FeatureGroupData(cache.get(group_name))
    feature_data: FeatureData = group_data.features.get(feature_name)
    group = group_data.map(name=group_name)
    handlers = [handler.map() for handler in feature_data.handlers]
    feature = feature_data.map(name=feature_name, group=group, handlers=handlers)
    return feature


from schematics import types as t, Model

# Container configuration


class ContainerConfiguration(Model):

    app_project_filepath = t.StringType(required=False, default=None)


# Default container
class Container():

    # Custom fields below
    # ...

    def __init__(self, config: ContainerConfiguration):
        # Default init
        self.config = config

        # Custom init below
        # ...

    def error_cache(self, flag: str = 'yaml'):
        if flag in ['yaml', 'yml']:
            from .repositories.error_cache.yaml import YamlErrorCache
            return YamlErrorCache(self.config.error_cache_path)
        
    def feature_cache(self, flag: str = 'yaml'):
        if flag in ['yaml', 'yml']:
            from .repositories.feature import yaml as feature_cache
            feature_cache.load_cache(self.config.feature_cache_path)
            return feature_cache


# Default dynamic container
class DynamicContainer():

    def add_service(self, service_name, factory_func):
        setattr(self, service_name, factory_func)



from ..objects import *


class FeatureHandlerData(FeatureHandler, DataObject):
    
    class Options(DefaultOptions):
        pass

    def map(self, role: str = 'to_object', **kwargs):
        return super().map(FeatureHandler, role, **kwargs)
    

class FeatureData(Feature, DataObject):

    class Options(DefaultOptions):
        roles = {
            'to_object.yaml': blacklist('handlers', 'group'),
            'to_data.yaml': blacklist('group')
        }

    handlers = t.ListType(t.ModelType(FeatureHandlerData), deserialize_from=['handlers', 'functions'])

    def map(self, role: str = 'to_object.yaml', **kwargs):
        return super().map(Feature, role, **kwargs)
    

class FeatureGroupData(FeatureGroup, DataObject):

    class Options(DefaultOptions):
        roles = {
            'to_object.yaml': blacklist('features'),
            'to_data.yaml': wholelist()
        }

    features = t.DictType(t.ModelType(FeatureData), default={})

    def map(self, role: str = 'to_object.yaml', **kwargs):
        return super().map(FeatureGroupData, role, **kwargs)


class AppError(Exception):

    def __init__(self, error_name: str, *format_args):
        super().__init__(self)
        self.error_name = error_name
        self.format_args = format_args


class DomainError(AppError):

    def __init__(self, error_name: str, *format_args):
        super().__init__(error_name, *format_args)


from schematics import types as t
from schematics import Model


class FeatureHandler(Model):
    name = t.StringType(required=True)
    function_path = t.StringType(required=True)
    data_mapping = t.StringType()
    use_services = t.StringType()
    params = t.DictType(t.StringType(), default={})
    log_activity = t.BooleanType(default=True)


class FeatureGroup(Model):
    name = t.StringType(required=True)
    data_mapping = t.StringType()


class Feature(Model):
    name = t.StringType(required=True)
    use_role = t.StringType()
    data_mapping = t.StringType()
    header_mapping = t.StringType()
    group = t.ModelType(FeatureGroup)
    handlers = t.ListType(t.ModelType(FeatureHandler), default=[])
    log_params = t.DictType(t.StringType(), default={})



from . import *

import yaml


class YamlErrorCache(ErrorCache):

    def __init__(self, cache_path: str):
        self.cache_path = cache_path


    def get(self, error_name: str, lang: str = 'en_US', error_type: type = Error) -> Error:
        with open(self.cache_path, 'r') as f:
            errors = yaml.safe_load(f)['errors']

        # First get the error data.
        try: 
            error_data = errors[error_name]
            message = error_data.pop('message')
            error: Error = error_type(dict(**error_data, error_name=error_name), strict=False)
        except KeyError:
            return None
        
        # Now try to get the error message according to the language.
        try:
            error.message = message[lang]
        except KeyError:
            # If the message for the regional language is not found, use the language instead.
            lang = lang.split('_')[0]
            error.message = message[lang]

        # Return the error.
        return error 


from .. import *
from . import *


def load_cache(cache_path: str):
    import yaml

    with open(cache_path, 'r') as f:
        data = yaml.safe_load(f)
        setattr(feature_cache, 'cache', data['features']['groups'])



from ...objects import *
from ...constants import *
from ...services import *
from ...errors import *

from schematics.exceptions import DataError


feature_cache = None
error_cache = None

def handle(feature_id: str, request, app_context, debug: bool = False, **kwargs):
    from ...error import ErrorManager
    from importlib import import_module
    from time import time
    
    class MessageContext():

        def __init__(self):
            self.headers = {}
            self.data = None
            self.services = None
            self.result = {}
            self.errors = ErrorManager()

    # Load feature cache.
    feature_cache = app_context.container.feature_cache()

    # Get feature from cache.
    feature: Feature = feature_cache.get(feature_id)
            
    # Create message context.
    context: MessageContext = MessageContext()

    # Add errors first.  It's easier this way...
    context.errors = app_context.errors

    # Begin header mapping process.
    if debug:
        print('Perform header mapping: "mapping": "{}"'.format(
            feature.header_mapping))

    # Import header mappings module.
    header_module = import_module(
        DEFAULT_HEADER_MAPPER_PATH.format(app_context.interface))

    try:
        # Retrieve header mapping function.
        header_mapping_func = getattr(
            header_module, feature.header_mapping)
    except TypeError:
        # Retrieve default header mapping function if none is specified.
        header_mapping_func = getattr(header_module, 'default')

    # Get header data and add to message context.
    context.headers = header_mapping_func(request, app_context, **kwargs)

    for handler in feature.handlers:

        if debug:
            print('Executing function: "function": "{}"'.format(
                function.to_primitive()))

        context.headers['request_start'] = int(time())

        # Set data mapping and service container for feature function
        try:
            data_mapping = feature_service.get_data_mapping(feature, handler)
            if debug: print('Perform data mapping: "mapping": "{}"'.format(data_mapping))
            data_mapping_func = getattr(
                import_module(DEFAULT_DATA_MAPPER_PATH.format(app_context.interface)),
                data_mapping)
            if debug: print('Performing data mapping for following request: "mapping": "{}", "request": "{}", params: "{}"'.format(data_mapping, request, handler.params))
            context.data = data_mapping_func(context, request, app_context, **handler.params, **kwargs)
            if debug: print('Data mapping complete: "mapping": "{}", "data": "{}"'.format(data_mapping, context.data.to_primitive()))
            # Request model state validation
            try:
                context.data.validate()
            except AttributeError: # In the case where there is no configured data mapping
                pass
        except TypeError as ex:
            print(ex)
        except DataError as ex:
            raise AppError(
                app_context.errors.INVALID_REQUEST_DATA.format_message(ex.messages))

        context.services = app_context.container

        # Format function module path.
        module_path = 'app.features.{}'.format(handler.function_path)

        # Import function module.
        if debug:
            print('Importing function: {}'.format(module_path))
        func = import_module(module_path)

        # This is a crazy test.
        if handler.function_path == 'care_recipients.assign_device':
            setattr(func, 'care_recipient_repo', context.services.care_recipient_repo())
            setattr(func, 'device_repo', context.services.device_repo())

        # Execute function handler.
        if debug:
            print('Executing function: {}'.format(module_path))
        result = func.handle(context)
        # For those who do now wish to assign the results to the context in the handler
        if result:
            context.result = result

        # Log activity
        if handler.log_activity:
            if debug:
                print('Logging activity for function: {}'.format(module_path))
            activity = import_module(DEFAULT_ACTIVITY_HANDLER_PATH)
            activity.handle(context)

        if debug:
            print('Finishing function: {}'.format(module_path))

    context.headers['request_end'] = int(time())

    # Return result
    # Handle list scenario
    if type(context.result) == list:
        result = []
        for item in context.result:
            if isinstance(item, Model):
                if feature.use_role:
                    result.append(item.to_primitive(
                        role=feature.use_role))
                else:
                    result.append(item.to_primitive())
            else:
                result.append(item)
        return result
    if not context.result:
        return {}
    # Convert schematics models to primitive dicts.
    if isinstance(context.result, Model):
        if feature.use_role:
            return context.result.to_primitive(role=feature.use_role)
        else:
            return context.result.to_primitive()
    return context.result
